// ===============================================
// File:		Objects.cxx
// Description:		VRML-Nodes defined with OpenGL
// Current:		Sept 28th 1999
// Author(s):		Thomas Kern
// ===============================================

/*
 *        thk VRML-Viewer (based on OpenGL)
 *  it shows static *.wrl files based on the VRML97 standard
 *
 *          Copyright (C) 1997-1999  Thomas Kern
 *                  All Rights Reserved
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Library General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free
 *   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * My email adress is: 
 *   Thomas.Kern@3Dsource.de 
 * 
 * You should find these files at: my homepage
 *   http://www.3Dsource.de
 *
 */

//=====================================================================
//	Description
//=====================================================================
// 
/*
  These objectes are used by VRMLControl
  Some funftions are located in Tools.h/.cxx.
*/

//=====================================================================
//	INCLUDES
//=====================================================================
//
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "Tools.H"

#include "Objects.H"

// bestimmte Funktionen lassen sich durch Extensions simulieren
#if !defined(GL_VERSION_1_1) && !defined(GL_VERSION_1_2)
#define glBindTexture	glBindTextureEXT
#endif


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Grouping-Node.
// Diese Klasse enthaelt Verweise auf andere Nodes (Childrens).
//=====================================================================
//
// Children hinzufuegen
void GroupingNode::AddChildren(VRMLNode *Children)
{
  if(MyChildrenCount < MAX_CHILDREN) {
    MyChildren[MyChildrenCount] = Children;
    MyChildrenCount++;
  }
  else
    writeMessage("GroupingNode::AddChildren: max Children erreicht.");
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Transform-Node.
// Diese Klasse erledigt die Transformationen (Rotation, Translation)
// und enthaelt Verweise auf die betroffenen Nodes.
//=====================================================================
//
// Zeichnen
void GroupNode::DrawNode()
{
  for(int i = 0; i < MyChildrenCount; i++)
    MyChildren[i]->DrawNode();
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Transform-Node.
// Diese Klasse erledigt die Transformationen (Rotation, Translation)
// und enthaelt Verweise auf die betroffenen Nodes.
//=====================================================================
//
// Konstruktor
TransformNode::TransformNode()
{
  center = rotation = scale = scaleOr = translation = false;
  myNodeType  = Transform;
  MyChildrenCount = 0;
}

//---------------------------------------------------------------------
// Center hinzufuegen
void TransformNode::Center(GLfloat Axle[3])
{
  for(int i = 0; i < 3; i++)
    CenM[i] = Axle[i];
  center = true;
}

//---------------------------------------------------------------------
// Rotation hinzufuegen
void TransformNode::Rotation(GLfloat Angle, GLfloat Axle[3])
{
  RotM[0] = Angle;
  if(Angle) {
    for(int i = 0; i < 3; i++)
      RotM[i+1] = Axle[i];
    rotation = true;
  }
}

//---------------------------------------------------------------------
// Scale hinzufuegen
void TransformNode::Scale(GLfloat Axle[3])
{
  for(int i = 0; i < 3; i++)
    ScaleM[i] = Axle[i];// Drehachse

  if( ScaleM[0] != 1.0 || ScaleM[1] != 1.0 || ScaleM[2] != 1.0 ) {
    scale = true;
    glEnable(GL_NORMALIZE);
  }
}

//---------------------------------------------------------------------
// ScaleOrientation hinzufuegen
void TransformNode::ScaleOrientation(GLfloat Angle, GLfloat Axle[3])
{
  ScaleOrM[0] = Angle;
  if(Angle) {
    for(int i = 0; i < 3; i++)
      ScaleOrM[i+1] = Axle[i];
    scaleOr = true;
  }
}

//---------------------------------------------------------------------
// Translation hinzufuegen
void TransformNode::Translation(GLfloat Axle[3])
{
  for(int i = 0; i < 3; i++)
    TransM[i] = Axle[i];// Drehachse
  translation = true;
}

//---------------------------------------------------------------------
// Zeichnen der Children
void TransformNode::DrawNode()
{
  int i;

  glPushMatrix();
  {
    // die Reihenfolge der Transformationen ist vorgeschrieben
    if(translation)
      glTranslatef( TransM[0], TransM[1], TransM[2] );
    if(center)
      glTranslatef( CenM[0], CenM[1], CenM[2] );
    if(rotation)
      glRotatef( RotM[0], RotM[1], RotM[2], RotM[3] );
    if(scale) {
      if(scaleOr)
	      glRotatef( ScaleOrM[0], ScaleOrM[1], ScaleOrM[2], ScaleOrM[3] );
      glScalef( ScaleM[0], ScaleM[1], ScaleM[2] );
      if(scaleOr)
	      glRotatef( -ScaleOrM[0], ScaleOrM[1], ScaleOrM[2], ScaleOrM[3] );
    }
    if(center)
      glTranslatef( -CenM[0], -CenM[1], -CenM[2] );

    // nun die Objekte
    for(i = 0; i < MyChildrenCount; i++)
      MyChildren[i]->DrawNode();
  }
  glPopMatrix();
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete TextureTransform-Node.
// Diese Klasse verzerrt das Texturbild eines Objektes.
//=====================================================================
//
// Konstruktor
TextureTransformNode::TextureTransformNode()
{
  center = rotation = scale = translation = false;
  myNodeType  = TextureTransform;
}

//---------------------------------------------------------------------
// Center hinzufuegen
void TextureTransformNode::Center(GLfloat Axle[2])
{
  for(int i = 0; i < 2; i++)
    CenM[i] = Axle[i];
  center = true;
}

//---------------------------------------------------------------------
// Rotation hinzufuegen
void TextureTransformNode::Rotation(GLfloat Angle)
{
  RotV = Angle;
  if(Angle)
    rotation = true;
}

//---------------------------------------------------------------------
// Scale hinzufuegen
void TextureTransformNode::Scale(GLfloat Axle[2])
{
  for(int i = 0; i < 2; i++)
    ScaleM[i] = Axle[i];

  if( ScaleM[0] != 1.0 || ScaleM[1] != 1.0 )
    scale = true;
}

//---------------------------------------------------------------------
// Translation hinzufuegen
void TextureTransformNode::Translation(GLfloat Axle[2])
{
  for(int i = 0; i < 2; i++)
    TransM[i] = Axle[i];
  translation = true;
}

//---------------------------------------------------------------------
// Veraendern des Matrix-Stacks
void TextureTransformNode::DrawNode()
{
  glMatrixMode(GL_TEXTURE);
  {
    glLoadIdentity();

    if(center)
      glTranslatef( -CenM[0], -CenM[1], 0.0 );
    if(scale)
      glScalef( ScaleM[0], ScaleM[1], 0.0 );
    if(rotation)
      glRotatef( RotV, 0.0, 0.0, 0.0 );
    if(center)
      glTranslatef( CenM[0], CenM[1], 0.0 );
    if(translation)
      glTranslatef( TransM[0], TransM[1], 0.0 );
  }
  glMatrixMode(GL_MODELVIEW);
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Texture-Node.
// Diese Klasse beschreibt das Texturbild eines Objektes.
//=====================================================================
//
// Konstruktor
TextureNode::TextureNode()
{
  textureID = 0;
  repeatS   = true;
  repeatT   = true;
  ImagePtr  = NULL;
  Breite    = Hoehe = 0;
  madeTexture = false;
  IsReady     = false;
  for(int i = 0; i < 4; i++)
    Color4[i] = 1.0;

  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
}

//---------------------------------------------------------------------
// Destruktor
TextureNode::~TextureNode()
{
  if(madeTexture)
    delete[] ImagePtr;
}

//---------------------------------------------------------------------
// Textur aufbauen (mit LOD-Stufen)
void TextureNode::PrepareTexture()
{
  if( madeTexture && (ImagePtr != NULL) ) {
    glBindTexture(GL_TEXTURE_2D, textureID);

    if(repeatS)            // Wiederholung der Textur
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    else
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);

    if(repeatT)            // Wiederholung der Textur
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    else
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    if(modulate)
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    else
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);

    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, Breite, Hoehe,
       	              GL_RGB, GL_UNSIGNED_BYTE, ImagePtr);    
    
    IsReady = true;
  }
}

//---------------------------------------------------------------------
// Textur aktivieren
void TextureNode::DrawNode()
{
  if( IsReady && drawWithTexture ) {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, textureID);
    // glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, Color4);
    if(modulate)
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    else
      glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  }
}


//=====================================================================
// Von der Texture-Node abgeleitete ImageTexture-Node.
//=====================================================================
//
// Konstruktor
ImageTextureNode::ImageTextureNode(int texturNr)
{
  textureID  = (unsigned int) texturNr;
  url[0] = '\0';
  myNodeType = ImageTexture;
}

//---------------------------------------------------------------------
// Bilddatei fuer die Textur einlesen
void ImageTextureNode::ScanTextureImage()
{
  enum {unknown, ppm, jpg}  FileType;
  const int        len3 = WORD_LENGHT;
  char             word[len3+1];     // \0-Zeichen muss auch noch rein
  char             testType[6];     // Filetyp feststellen
  int              i;
  FILE            *Input  = NULL;
  
  if( CheckFile(url, "r") )
    Input = fopen(url, "r");

  if( Input == NULL  ) {
    sprintf(myMessageString,"Textur '%s' nicht gefunden.", url);
    showMessage();
  }
  else {
    // den Dateityp feststellen
    FileType = unknown;
    fread(word, 1, 10, Input);  // die ersten 10 Bytes lesen
    testType[0] = word[0];     // auf PPM die ersten beiden Bytes 
    testType[1] = word[1];     // pruefen
    testType[2] = '\0';
    if( !strcmp(testType, "P3") || !strcmp(testType, "P6") )
      FileType = ppm;
    else {
      for(i = 0; i < 4; i++)
        testType[i] = word[i+6];
        testType[4] = '\0';            
      if( !strcmp(testType, "JFIF") )
        FileType = jpg;
    }
    
    //nicht vergessen, die Datei zu schliessen
    fclose(Input);

    // nun das Laden der Datei veranlassen
    ImagePtr = NULL;
    switch(FileType) {
    case ppm:
      ImagePtr = loadPPMfile(url, Breite, Hoehe);
      break;
    case jpg:
      ImagePtr = loadJPEGfile(url, Breite, Hoehe);
      break;
    default:
      showMessage("Image-Format wird nicht unterstuetzt.");
      break;
    }
    if(ImagePtr)
      madeTexture = true;
    else       
      madeTexture = false;
  }
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Material-Node.
// Diese Klasse beschreibt die Materialeigenschaften von Objekten.
//=====================================================================
//
// Konstruktor
MaterialNode::MaterialNode()
{
  int i;

  AmbientMult = 0.2;

  for(i = 0; i < 3; i++) {
    DiffuseColor4[i] = 0.8;
    AmbientColor4[i] = AmbientMult*DiffuseColor4[i];
  }
  AmbientColor4[3] = DiffuseColor4[3] = 1.0; // Transparenz

  useMaterial = false;  // nur die Farbwerte nutzen (ohne Material)
  myNodeType  = Material;
}

//---------------------------------------------------------------------
// Farbe unter ambientem Licht (Umgebungslicht von allen Seiten)
void MaterialNode::AddAmbient(GLclampf Intensity)
{
  AmbientMult = Intensity;
  for(int i = 0; i < 3; i++)
    AmbientColor4[i] = AmbientMult*DiffuseColor4[i];
}

//---------------------------------------------------------------------
// Farbe unter diffusem Licht (Licht aus einer bestimmten Richtung)
void MaterialNode::AddDiffuse(GLclampf Color[3])
{
  for(int i = 0; i < 3; i++) {
    DiffuseColor4[i] = Color[i];
    AmbientColor4[i] = AmbientMult*DiffuseColor4[i];
  }
}

//---------------------------------------------------------------------
// Transparenz des Materials
void MaterialNode::AddTransparency(GLclampf Intensity)
{
  // In der OpenGL bedeutet 1.0 == undurchsichtig. Die VRML97-
  // Spec legt das Gegenteil fest.
  AmbientColor4[3] = DiffuseColor4[3] = (1.0 - Intensity); // Transparenz
}

//---------------------------------------------------------------------
// Status setzen
void MaterialNode::DrawNode()
{
  if(useMaterial) {
    glMaterialfv(GL_FRONT, GL_AMBIENT, AmbientColor4);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, DiffuseColor4);
  }
  else
    glColor4fv(DiffuseColor4);
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Appearance-Node.
// Diese Klasse beschreibt das Aussehen von Objekten.
//=====================================================================
//
// Materialeigenschaften setzen
void AppearanceNode::DrawNode()
{
  if( haveMaterial )
    MyMaterial->DrawNode();
  else {
    MyMaterial = NULL;
    MyMaterial = new MaterialNode();
    if(MyMaterial != NULL)
      haveMaterial = ownMaterial = true;
    MyMaterial->DrawNode();
  }
  if( haveTexture ) {
    if( haveTextureTransform )
      MyTextureTransform->DrawNode();
    else {
      MyTextureTransform = NULL;
      MyTextureTransform = new TextureTransformNode();
      if(MyTextureTransform != NULL)
        haveTextureTransform = ownTextureTransform = true;
      MyTextureTransform->DrawNode();
    }
    MyTexture->DrawNode();
  }
}


//=====================================================================
// Von der allgemeinen VRML-Node abgeleitete Field-Node.
// Diese Klasse stellt ein Datenfeld float-Werten zur Verfuegung.
// Erzeugt wird das Datenfeld vom Aufrufer direkt aus einer Datei.
// Diese Klasse kann fuer die Nodes: Color, Coordinate, Normal
// und TextureCoordinate angewendet werden. Zur eindeutigen Zuordnung
// sollte der Typ explizit gesetzt werden (SetNodeTyp).
//=====================================================================
//
// Konstruktor
FieldNode::FieldNode()
{
#ifdef USE_STATIC_FIELDS
  VertexCount  = 3*MAXVERTICES;
#else
  myValueField = NULL;
  VertexCount  = 0;
#endif
}

//---------------------------------------------------------------------
// Destruktor
FieldNode::~FieldNode()
{
#ifndef USE_STATIC_FIELDS
  if( myValueField != NULL )
    delete myValueField;
#endif
}

//---------------------------------------------------------------------
// ausreichend grosses Feld erzeugen und Werte kopieren
void FieldNode::NewField(GLfloat *FieldPtr, int FieldSize)
{
  VertexCount = FieldSize;
#ifndef USE_STATIC_FIELDS
  myValueField = NULL;
  if(VertexCount) {
    myValueField = new GLfloat[VertexCount];
    if(myValueField != NULL)
      memcpy( myValueField, FieldPtr, VertexCount*sizeof(GLfloat) );
  }
#else
  if(VertexCount)
    memcpy( myValueField, FieldPtr, VertexCount*sizeof(GLfloat) );
#endif
}


//=====================================================================
// Von einem VRML-Object abgeleitete Box-Node.
// Diese Klasse beschreibt einen Quader.
//=====================================================================
//
// Konstruktor
BoxNode::BoxNode()
{
  GLfloat   BoxM[3];
  BoxM[0] = BoxM[1] = BoxM[2] = 2.0;
  AddSize(BoxM);

  myNodeType = Box;
}

//---------------------------------------------------------------------
// Geometriedaten aendern
void BoxNode::AddSize(GLfloat GeomMatrix[3])
{
  // Eckpunkte berechnen, Koerper ist symmetrisch zum Ursprung
  GLfloat   x = 0.5*GeomMatrix[0];
  GLfloat   y = 0.5*GeomMatrix[1];
  GLfloat   z = 0.5*GeomMatrix[2];

  // separate Arrays => Zeiger muessen nicht aufgeloest werden
  Plu[0] =   -x;   Plu[1] =   -y;   Plu[2] =   z; // vorn, links unten
  Pru[0] =    x;   Pru[1] =   -y;   Pru[2] =   z; //       rechts unten
  Pro[0] =    x;   Pro[1] =    y;   Pro[2] =   z; //       rechts oben
  Plo[0] =   -x;   Plo[1] =    y;   Plo[2] =   z; //       links oben
  Qlu[0] =   -x;   Qlu[1] =   -y;   Qlu[2] =  -z; // hinten, links unten
  Qru[0] =    x;   Qru[1] =   -y;   Qru[2] =  -z; //         rechts unten
  Qro[0] =    x;   Qro[1] =    y;   Qro[2] =  -z; //         rechts oben
  Qlo[0] =   -x;   Qlo[1] =    y;   Qlo[2] =  -z; //         links oben

  // Normalenvektoren berechnen
  CalcNormal(Plu, Pru, Plo, Nfl[0]);  // vorn
  CalcNormal(Pru, Qru, Qro, Nfl[1]);  // rechts
  CalcNormal(Qru, Qlu, Qro, Nfl[2]);  // hinten
  CalcNormal(Qlu, Plu, Qlo, Nfl[3]);  // links
  CalcNormal(Qlu, Qru, Plu, Nfl[4]);  // unten
  CalcNormal(Plo, Pro, Qro, Nfl[5]);  // oben

  Tlu[0] =  0.0;   Tlu[1] =  0.0;                 // Textur-Eckpunkte 2D
  Tru[0] =  1.0;   Tru[1] =  0.0;
  Tro[0] =  1.0;   Tro[1] =  1.0;
  Tlo[0] =  0.0;   Tlo[1] =  1.0;
}

//---------------------------------------------------------------------
// Zeichne die reine Geometrie
void BoxNode::DrawNode()
{
  if(haveAppearance)
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
  else
    glColor3f(0.8, 0.8, 0.8);            // Standardfarbe

  if(drawWithTexture)
    DrawWithTexture();
  else {
    glBegin(GL_QUAD_STRIP);      // (entgegen dem Uhrzeigersinn)
    {
      glNormal3fv(Nfl[0]);
      glVertex3fv(Plu);          // 1 vorn
      glVertex3fv(Pru);          // 2
      glVertex3fv(Plo);          // 3
      glVertex3fv(Pro);          // 4
      glNormal3fv(Nfl[5]);
      glVertex3fv(Qlo);          // 5 oben:   mit 3,4
      glVertex3fv(Qro);          // 6
      glNormal3fv(Nfl[2]);
      glVertex3fv(Qlu);	       // 7 hinten: mit 5,6
      glVertex3fv(Qru);          // 8
      glNormal3fv(Nfl[4]);
      glVertex3fv(Plu);	       // 9 unten:  mit 7,8
      glVertex3fv(Pru);          //10
    }
    glEnd();
    glBegin(GL_QUADS);
    {
      glNormal3fv(Nfl[3]);
      glVertex3fv(Plu);          // 1 links
      glVertex3fv(Plo);          // 2
      glVertex3fv(Qlo);          // 3
      glVertex3fv(Qlu);          // 4
      glNormal3fv(Nfl[1]);
      glVertex3fv(Qru);	       // 1 rechts
      glVertex3fv(Qro);          // 2
      glVertex3fv(Pro);          // 3
      glVertex3fv(Pru);          // 4
    }
    glEnd();
  }
}

//---------------------------------------------------------------------
// Geometrie und Textur
void BoxNode::DrawWithTexture()
{
//  glColor3f(1.0, 1.0, 1.0);

  glBegin(GL_QUAD_STRIP);      // (entgegen dem Uhrzeigersinn)
  {
    glNormal3fv(Nfl[0]);
    glTexCoord2fv(Tlu);	glVertex3fv(Plu);  // 1 vorn
    glTexCoord2fv(Tru); glVertex3fv(Pru); // 2
    glTexCoord2fv(Tlo); glVertex3fv(Plo); // 3
    glTexCoord2fv(Tro);	glVertex3fv(Pro);  // 4
    glNormal3fv(Nfl[5]);
    glTexCoord2fv(Tlo);	glVertex3fv(Qlo);  // 5 oben:   mit 3,4
    glTexCoord2fv(Tro);	glVertex3fv(Qro);  // 6
    glNormal3fv(Nfl[2]);
    glTexCoord2fv(Tlo);	glVertex3fv(Qlu);  // 7 hinten: mit 5,6
    glTexCoord2fv(Tro);	glVertex3fv(Qru);  // 8
    glNormal3fv(Nfl[4]);
    glTexCoord2fv(Tlo);	glVertex3fv(Plu);  // 9 unten:  mit 7,8
    glTexCoord2fv(Tro);	glVertex3fv(Pru);  //10
  }
  glEnd();
  glBegin(GL_QUADS);
  {
    glNormal3fv(Nfl[3]);
    glTexCoord2fv(Tru);	glVertex3fv(Plu);  // 1 links
    glTexCoord2fv(Tro);	glVertex3fv(Plo);  // 2
    glTexCoord2fv(Tlo);	glVertex3fv(Qlo);  // 3
    glTexCoord2fv(Tlu); glVertex3fv(Qlu);  // 4
    glNormal3fv(Nfl[1]);
    glTexCoord2fv(Tru); glVertex3fv(Qru);  // 1 rechts
    glTexCoord2fv(Tro); glVertex3fv(Qro);  // 2
    glTexCoord2fv(Tlo);	glVertex3fv(Pro);  // 3
    glTexCoord2fv(Tlu); glVertex3fv(Pru);  // 4
  }
  glEnd();
}


//=====================================================================
// Von einem VRML-Object abgeleitete Cone-Node.
// Diese Klasse beschreibt einen Kegel.
//=====================================================================
//
// Konstruktor
ConeNode::ConeNode()
{
  R = 2.0;
  H = 1.0;
  CalcGeometry();

  drawSide   = true;
  drawBottom = true;
  myNodeType  = Cone;
}

//---------------------------------------------------------------------
// Geometrie berechnen
void ConeNode::CalcGeometry()
{
  int      	i;
  GLfloat	dalpha;

  // Geometrie
  split  = MAX_ECKEN;               // Kreis aus n-Ecken (Polygonen)
  dalpha = (M_PI/180)*(360/split);  // Winkel zw. 2 Punkten (x,z-Ebene)

  P[0][0] = R;   P[0][1] = 0.0;   P[0][2] = 0.0;
  for(i = 1; i < split; i++) {
    P[i][0] =  R*cos(i*dalpha);
    P[i][1] =  0.0;
    P[i][2] = -R*sin(i*dalpha);
  }
  Q[0] = 0.0;   Q[1] = H;     Q[2] = 0.0;
  U[0] = 0.0;   U[1] = 0.0;   U[2] = 0.0;

  // Normalenvektoren berechnen
  for(i = 0; i < split-1; i++)
    CalcNormal(P[i], P[i+1], Q, Nfl[i]);
  CalcNormal(P[split-1], P[0], Q, Nfl[split-1]);    // Seite
  
  CalcNormal(P[0], U, P[split-1], Nfl[split]);      // Boden
  
}

//---------------------------------------------------------------------
// Zeichnung ausfuehren
void ConeNode::DrawNode()
{
  int i;

  if(haveAppearance)
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
  else
    glColor3f(0.8, 0.8, 0.8);            // Standardfarbe
  

  if(drawSide) {                       // Mantelflaeche zeichnen
    glBegin(GL_TRIANGLE_FAN);
    {
      glNormal3fv(Nfl[0]);
      glVertex3fv(Q);
      for(i = 0; i < split-1; i++) {
	      glNormal3fv(Nfl[i+1]);
        glVertex3fv(P[i]);
      }
      glVertex3fv(P[split-1]);
      glVertex3fv(P[0]);
    }
    glEnd();
  }

  if(drawBottom) {                     // Bodenflaeche zeichnen (CCW !)
    glBegin(GL_TRIANGLE_FAN);
    {
      glNormal3fv(Nfl[split]);
      glVertex3fv(U);
      glVertex3fv(P[0]);
      for(i = split-1; i >= 0; i--)
	      glVertex3fv(P[i]);
    }
    glEnd();
  }
}


//=====================================================================
// Von einem VRML-Object abgeleitete Cylinder-Node.
// Diese Klasse beschreibt einen Zylinder.
//=====================================================================
//
CylinderNode::CylinderNode()
{
  R = 2.0;
  H = 1.0;
  CalcGeometry();

  drawSide   = true;
  drawBottom = true;
  drawTop    = true;
  myNodeType  = Cylinder;
}

//---------------------------------------------------------------------
// Geometrie berechnen
void CylinderNode::CalcGeometry()
{
  int     i;
  GLfloat dalpha;
  
  // Geometrie
  split  = MAX_ECKEN;  // Kreis aus n=MAX_ECKEN+1 Polygone
  dalpha = (M_PI/180)*(360/split);  // Winkel zw. 2 Punkten (x,z-Ebene)

  P[0][0] = R;  P[0][1] = -H; P[0][2] = 0.0;
  Q[0][0] = R;  Q[0][1] =  H; Q[0][2] = 0.0;
  UP[0] = 0.0;  UP[1] = -H;   UP[2] = 0.0;
  UQ[0] = 0.0;  UQ[1] =  H;   UQ[2] = 0.0;
  for(i = 1; i < MAX_ECKEN; i++) {
    P[i][0] =      Q[i][0] =  R*cos(i*dalpha);
    P[i][1] = -H;  Q[i][1] =  H;
    P[i][2] =      Q[i][2] = -R*sin(i*dalpha);
  }
  // Normalenvektoren berechnen
  for(i = 0; i < split-1; i++)
    CalcNormal(Q[i], P[i], Q[i+1], Nfl[i]);
  CalcNormal(Q[split-1], P[split-1], Q[0], Nfl[split-1]); // Seite

  CalcNormal(P[1], P[0], UP, Nfl[split]);      // Boden
  CalcNormal(Q[0], Q[1], UQ, Nfl[split+1]);    // Deckel

}

//---------------------------------------------------------------------
// Zeichnung ausfuehren
void CylinderNode::DrawNode()
{
  int i;

  if(haveAppearance) {
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
    // if(drawWithTexture)
    // DrawWithTexture();
  }
  else
    glColor3f(0.8, 0.8, 0.8);            // Standardfarbe

  if(drawBottom) {                   // Bodenflaeche zeichnen (CCW !)
    glBegin(GL_TRIANGLE_FAN);
    {
      glNormal3fv(Nfl[split]);
      glVertex3fv(UP);
      glVertex3fv(P[0]);
      for(i = split-1; i >= 0; i--)
	      glVertex3fv(P[i]);
    }
    glEnd();
  }

  if(drawSide) {                     // Mantelflaeche zeichnen
    glBegin(GL_QUAD_STRIP);
    {
      glNormal3fv(Nfl[0]);
      glVertex3fv(Q[0]);
      glVertex3fv(P[0]);
      glVertex3fv(Q[1]);
      glVertex3fv(P[1]);
      for(i = 2; i < split; i++) {
	      glNormal3fv(Nfl[i-1]);
	      glVertex3fv(Q[i]);
	      glVertex3fv(P[i]);
      }
      glNormal3fv(Nfl[split-1]);
      glVertex3fv(Q[0]);
      glVertex3fv(P[0]);
    }
    glEnd();
  }

  if(drawTop) {                      // Deckel zeichnen (CW !)
    glBegin(GL_TRIANGLE_FAN);
    {
      glNormal3fv(Nfl[split+1]);
      glVertex3fv(UQ);
      for(i = 0; i < split; i++)
	      glVertex3fv(Q[i]);
      glVertex3fv(Q[0]);
    }
    glEnd();
  }
}


//=====================================================================
// Von einem VRML-Object abgeleitete Sphere-Node.
// Diese Klasse beschreibt eine Kugel.
//=====================================================================
//
// Konstruktor
SphereNode::SphereNode()
{
  Quadric = gluNewQuadric();

  R = 1.0;
  CalcGeometry();

  myNodeType  = Sphere;
}

//---------------------------------------------------------------------
// Destruktor
SphereNode::~SphereNode()
{
  gluDeleteQuadric(Quadric);
}

//---------------------------------------------------------------------
// Geometrie berechnen
void SphereNode::CalcGeometry()
{
  split = (int)(0.75*MAX_ECKEN);
}

//---------------------------------------------------------------------
// Zeichnung ausfuehren
void SphereNode::DrawNode()
{
  if(haveAppearance) {
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
    if(drawWithTexture) {
      //glColor3f(1.0, 1.0, 1.0); // weisser Hintergrund
      gluQuadricTexture(Quadric, GL_TRUE);
    }
  }
  else
    glColor3f(0.8, 0.8, 0.8);            // Standardfarbe

  gluSphere( Quadric, R, split, split );
}

//=====================================================================
// Von einem VRML-Object abgeleitete FaceSet-Node.
// Diese Klasse beschreibt einen Polygonzug.
//=====================================================================
//
// Konstruktor
FaceSetNode::FaceSetNode()
{
  myDrawing	       = nothingToDraw;
  whichDrawing     = (int)myDrawing;
#ifdef USE_STATIC_FIELDS
   MyCoordIndexSize  = MyColorIndexSize   = INDEX_MULT*MAXVERTICES;
   MyTextureIndexSize = INDEX_MULT*MAXVERTICES;
#else
  MyCoordIndex = MyColorIndex = MyTextureIndex = NULL;
  MyCoordIndexSize = MyColorIndexSize   = 0;
  MyTextureIndexSize = 0;
#endif
  haveCoords       = haveColors = false;
  haveNormals      = haveTextureCoords = false;
  haveCoordIndex   = haveColorIndex    = false;
  haveTextureIndex = false;
  haveOwnNormals   = false;
  CCW = ColorPerVertex = Convex = NormalPerVertex = true;
  creaseAngle	     = 0.0;

  myNodeType       = IndexedFaceSet;
}

//---------------------------------------------------------------------
// Destruktor
FaceSetNode::~FaceSetNode()
{
#ifndef  USE_STATIC_FIELDS
  if( MyCoordIndex != NULL )
    delete MyCoordIndex;
  if( MyColorIndex != NULL )
    delete MyColorIndex;
  if( MyTextureIndex != NULL )
    delete MyTextureIndex;
#endif
  if(haveOwnNormals)
    delete MyNormals;
}

//---------------------------------------------------------------------
// Zeiger auf Koordinatenfeld
void FaceSetNode::AddCoordField(GLfloat *CoordField)
{
  MyCoords   = CoordField;
  haveCoords = true;
}

//---------------------------------------------------------------------
// Indizes auf Koordinatenfeld
void FaceSetNode::AddCoordIndex(int *CoordIndexPtr, int CoordIndexSize)
{
#ifdef  USE_STATIC_FIELDS
  memcpy( MyCoordIndex, CoordIndexPtr, CoordIndexSize*sizeof(int) );
  MyCoordIndexSize = CoordIndexSize;
  haveCoordIndex   = true;
#else
  MyCoordIndex = NULL;
  MyCoordIndex = new int[CoordIndexSize];
  if(MyCoordIndex != NULL) {
    memcpy( MyCoordIndex, CoordIndexPtr, CoordIndexSize*sizeof(int) );
    MyCoordIndexSize = CoordIndexSize;
    haveCoordIndex   = true;
  }
#endif
}

//---------------------------------------------------------------------
// Zeiger auf ColorFeld
void FaceSetNode::AddColorField(GLfloat *ColorField)
{
  MyColors   = ColorField;
  haveColors = true;
}

//---------------------------------------------------------------------
// Indizes auf ColorFeld
void FaceSetNode::AddColorIndex(int *ColorIndex, int ColorIndexSize)
{
#ifdef  USE_STATIC_FIELDS
  memcpy( MyColorIndex, ColorIndex, ColorIndexSize*sizeof(int) );
  MyColorIndexSize = ColorIndexSize;
  haveColorIndex   = true;
#else
  MyColorIndex = NULL;
  MyColorIndex = new int[ColorIndexSize];
  if(MyColorIndex != NULL) {
    memcpy( MyColorIndex, ColorIndex, ColorIndexSize*sizeof(int) );
    MyColorIndexSize = ColorIndexSize;
    haveColorIndex   = true;
  }
#endif
}

//---------------------------------------------------------------------
// Zeiger auf Normalenvektor-Feld
void FaceSetNode::AddNormalField(GLfloat *NormalField)
{
  MyNormals   = NormalField;
  haveNormals = true;
}

//---------------------------------------------------------------------
// Zeiger auf Texturkoordinatem-Feld
void FaceSetNode::AddTextureField(GLfloat *TextureCoordField)
{
  MyTextureCoords   = TextureCoordField;
  haveTextureCoords = true;
}

//---------------------------------------------------------------------
// Indizes auf Texturkoordinatem-Feld
void FaceSetNode::AddTextureIndex(int *TextureCoordIndex, int TextureIndexSize)
{
#ifdef  USE_STATIC_FIELDS
  memcpy( MyTextureIndex, TextureCoordIndex, TextureIndexSize*sizeof(int) );
  MyTextureIndexSize = TextureIndexSize;
  haveTextureIndex   = true;
#else
  MyTextureIndex = NULL;
  MyTextureIndex = new int[TextureIndexSize];
  if(MyTextureIndex != NULL) {
    memcpy( MyTextureIndex, TextureCoordIndex, TextureIndexSize*sizeof(int) );
    MyTextureIndexSize = TextureIndexSize;
    haveTextureIndex   = true;
  }
#endif
}

//---------------------------------------------------------------------
// Normalenvektoren berechnen
void FaceSetNode::CalcNormals()
{
  int       i, j, IndexCount, oldIndexCount, VerticesPerFace;
  int       PointCount = 0;
  GLdouble  FaceNormal[3], nbetrag, kehrwert; // mittlerer NV und Betrag
  GLdouble  cosWinkel;
  const int maxNVperPoint = 50; 	     // max. Normalenvektoren pro Punkt
  int       IndexNr[maxNVperPoint];	     // Position der Punkte im Indexfeld
  int       k, indexPos;		     // weitere Zaehler
  GLdouble  smoothNormal[3];                 // mittlerer NV
  GLdouble  NVwinkel[3], winkel, a, b;

  // Datenfeld so gross waehlen, dass ein Punkt auch mehrere
  // Normalenvektoren (fuer jede Flaeche einen) haben kann.
  // Die Berechnung erfolgt so, dass die Normalenvektoren in der
  // Reihenfolge der genutzten Punkte bereitgestellt werden, also
  // nicht ueber spezielle Indizes !
  MyNormals = NULL;
  int mindestgroesse = MyCoordIndexSize;      // falls nur wenig Speicher
  if( mindestgroesse < 120 )		      // angefordert wird, erfolgt ein
    mindestgroesse = 120;	              // regelmaessiger Absturz !!
  MyNormals = new GLfloat[3*mindestgroesse];  // Datenfeld anlegen (XYZ)
  if(MyNormals != NULL)
    haveOwnNormals = true;
  else
    return;                                    // den restlichen Block ueberspringen

  // nun fuer jeden Eckpunkt der Flaechen die Normalenvektoren bestimmen
  IndexCount = 0;
  while( IndexCount < MyCoordIndexSize -3 ) {  // mindestens 3 Vertices
    oldIndexCount   = IndexCount;              // aktueller Startpunkt
    VerticesPerFace = 0;                       // noch nicht zu Flaechen zugeordnet
    while( (MyCoordIndex[IndexCount] >= 0) &&  // -1 ist Abschluss
    	   (IndexCount < MyCoordIndexSize) ) {
      VerticesPerFace++;                       // Anzahl der Eckpunkte dieser Flaeche
      if( MyCoordIndex[IndexCount] > PointCount )
	      PointCount++;                          // fuer Smoothing
      IndexCount++;
    }
    i = 1;
    while( i < VerticesPerFace-1 ) {           // ersten und letzten getrennt
      // es werden Zeiger auf ein float-Feld uebergeben (...)
      CalcNormal( (MyCoords  + 3*MyCoordIndex[oldIndexCount + i    ] ),// Mitte
		              (MyCoords  + 3*MyCoordIndex[oldIndexCount + i + 1] ),// rechts
		              (MyCoords  + 3*MyCoordIndex[oldIndexCount + i - 1] ),// links
		              (MyNormals + 3*(oldIndexCount+i) ) );                // Normal
      i++;
    }
    // Normalenvektor erster Eckpunkt
    i = VerticesPerFace -1;          // bezeichnet den letzten Eckpunkt
    CalcNormal( (MyCoords  + 3*MyCoordIndex[oldIndexCount    ] ),// Mitte
		            (MyCoords  + 3*MyCoordIndex[oldIndexCount + 1] ),// rechts
		            (MyCoords  + 3*MyCoordIndex[oldIndexCount + i] ),// links
		            (MyNormals + 3*oldIndexCount ) );               // Normal
    // Normalenvektor letzter Eckpunkt
    CalcNormal( (MyCoords  + 3*MyCoordIndex[oldIndexCount + i    ] ),// Mitte
		            (MyCoords  + 3*MyCoordIndex[oldIndexCount        ] ),// rechts
		            (MyCoords  + 3*MyCoordIndex[oldIndexCount + i - 1] ),// links
		            (MyNormals + 3*(oldIndexCount +i) ) );              // Normal

    IndexCount++;				 // die -1 ueberspringen
  }

#ifdef THK_TEST_DRUCK
  // Testausdruck
  i = 0; j = 0;
  while( i < MyCoordIndexSize ) {
    printf("Face %d\n", j);
    j++;
    while( (MyCoordIndex[i] >= 0) && ( i < MyCoordIndexSize) ) {
      printf("Normal before Change: %2.2f, %2.2f, %2.2f\n",
             MyNormals[3*i], MyNormals[3*i+1], MyNormals[3*i+2] );
      i++;
    }
    i++;
  }
#endif // THK_TEST_DRUCK

  // Bei CW-Orientierung der Polygone zeigen die NV in die
  // falsche Richtung.
  if(!CCW) {
    for(i = 0; i < MyCoordIndexSize; i++)
      for(j = 0; j < 3; j++)
        MyNormals[3*i+j] *= -1;
  }

  // Da an dieser Stelle auch noch konkave Polygone existieren koennen,
  // mu die Richtung der NV noch auf ihre Plausibilitaet hin untersucht
  // werden. Dazu wird der mittlere NV der Flaeche berechnet. Weicht ein
  // NV um mehr als 90 Grad hiervon ab, so wird er umgedreht.
#define FLIP_NORMALS
#ifdef FLIP_NORMALS
  if(!Convex) {
    i = 0;
    while( i < MyCoordIndexSize ) {
      j = 0;
      FaceNormal[0] = FaceNormal[1] = FaceNormal[2] = 0.0;
      while( (MyCoordIndex[i+j] >= 0) && ( (i+j) < MyCoordIndexSize) ) {
        FaceNormal[0] += (GLdouble)MyNormals[3*i+j   ];
        FaceNormal[1] += (GLdouble)MyNormals[3*i+j +1];
        FaceNormal[2] += (GLdouble)MyNormals[3*i+j +2];
        j++;
      }
      nbetrag = sqrt( FaceNormal[0]*FaceNormal[0] +
	       	            FaceNormal[1]*FaceNormal[1] +
                      FaceNormal[2]*FaceNormal[2] );
      if(nbetrag) {
        kehrwert = 1/nbetrag;
        for(j = 0; j < 3; j++)
          FaceNormal[j] *= kehrwert;

        while( (MyCoordIndex[i] >= 0) && ( i < MyCoordIndexSize) ) {
          cosWinkel = ( FaceNormal[0]*MyNormals[3*i   ] +
	    	                FaceNormal[1]*MyNormals[3*i +1] +
                        FaceNormal[2]*MyNormals[3*i +2]   );
	        if(cosWinkel < 0.0) {
            for(j = 0; j < 3; j++)
              MyNormals[3*i+j] *= -1.0;
          }
          i++;
        }
      }
      i++;      			     // das ist die -1
    }
  }
#endif // FLIP_NORMALS

#ifdef THK_TEST_DRUCK
  // Testausdruck
  i = 0; j = 0;
  while( i < MyCoordIndexSize ) {
    printf("Face %d\n", j);
    j++;
    while( (MyCoordIndex[i] >= 0) && ( i < MyCoordIndexSize) ) {
      printf("Normal after Change: %2.2f, %2.2f, %2.2f\n",
             MyNormals[3*i], MyNormals[3*i+1], MyNormals[3*i+2] );
      i++;
    }
    i++;
  }
#endif // THK_TEST_DRUCK


  // Die VRML97-Spec sieht vor, dass die Normalenvektoren genau dann
  // aufeinander abgestimmt werden muessen, wenn der Winkel zwischen den
  // Normalenvektoren zweier aneinander grenzender Flaechen kleiner ist
  // als der creaseAngle [rad]. So erhaelt man einen weichen Verlauf der
  // Schattierung. Da ein Punkt prinzipiell beliebig vielen Flaechen
  // zugeordnet werden kann, ist die Berechnung entsprechend aufwendig.
  // Genaugenommen muessten alle Normalenvektoren eines Punktes (denn jeder
  // gemeinsame Eckpunkt hat auch fuer jede Flaeche einen eigenen NV) jeweils
  // miteinander verglichen werden.
  // Auf diesen Vergleich wird hier verzichtet. Statt dessen wird durch
  // Summierung aller Normalenvektoren der gemeinsame mittlere NV berechnet,
  // welcher dann gegen creaseAngle mit den einzelnen NV verglichen wird.
  // Bei zu grosser Abweichung ersetzt der mittlere NV den einzelnen NV.
  if(creaseAngle) {                        // Smoothing
    for(i = 0; i < PointCount; i++) {      // groesster Verweis auf Punkt
      for(j = 0; j < MyCoordIndexSize; j++) { // passenden Eintrag im Indexfeld suchen
        if( MyCoordIndex[j] == i ) {       // falls gefunden
    	    k = j;                           // Position ist Startpunkt
          indexPos = 0;			   // Anzahl der gleichen Index-Eintraege
	        while( k < MyCoordIndexSize ) {  // restliches Feld durchsuchen
            if( MyCoordIndex[k] == i &&    // den gleichen Index gefunden
                indexPos < maxNVperPoint ) {// Feldgroesse ist begrenzt
              IndexNr[indexPos] = k;       // Position merken
              indexPos++;
            }
            k++; 			   // alle Uebereinstimmungen suchen
          }
          if(indexPos > 1) {               // Punkt gehoert zu mehreren Flaechen
	          smoothNormal[0] = smoothNormal[1] = smoothNormal[2] = 0.0;
	          for(k = 0; k < indexPos; k++) {// mittleren NV bestimmen
              smoothNormal[0] += MyNormals[ 3*IndexNr[k]    ];
              smoothNormal[1] += MyNormals[ 3*IndexNr[k] + 1];
              smoothNormal[2] += MyNormals[ 3*IndexNr[k] + 2];
            }
	          nbetrag = sqrt( smoothNormal[0]*smoothNormal[0] +
	                   		    smoothNormal[1]*smoothNormal[1] +
                            smoothNormal[2]*smoothNormal[2] );
            if(nbetrag) {
              kehrwert = 1/nbetrag;
              for(k = 0; k < 3; k++)
                smoothNormal[k] *= kehrwert;
              // Betrag des Winkels zwischen Achse und mittlerem NV
              for(k = 0; k < 2; k++) {     // X und Y-Achse, mathematisch positiv
                if(smoothNormal[k])
                  NVwinkel[k] = atan( smoothNormal[k+1]/smoothNormal[k]);
                else
                  NVwinkel[k] = sign(smoothNormal[k+1])*M_PI_2;
              }
              if(smoothNormal[2])          // noch Z->X-Achse
                NVwinkel[2] = atan( smoothNormal[0]/smoothNormal[2]);
              else
                NVwinkel[2] = sign(smoothNormal[0])*M_PI_2;

              // jetzt Winkel zu jeder Achse vergleichen
              // und bei Bedarf anpassen
	            for(k = 0; k < indexPos; k++) {
                a = MyNormals[ 3*IndexNr[k] + 1]; // Y-Value
                b = MyNormals[ 3*IndexNr[k]    ]; // X-Value
                if(b)
                  winkel = atan(a/b);      // Winkel zur X-Achse
                else
                  winkel = sign(a)*M_PI_2;
                if( fabs(winkel - NVwinkel[0]) < creaseAngle )
                  MyNormals[ 3*IndexNr[k] ] = smoothNormal[0];
                a = MyNormals[ 3*IndexNr[k] + 2]; // Z-Value
                b = MyNormals[ 3*IndexNr[k] + 1]; // Y-Value
		            if(b)
                  winkel = atan(a/b);      // Winkel zur Y-Achse
                else
		              winkel = sign(a)*M_PI_2;
                if( fabs(winkel - NVwinkel[1]) < creaseAngle )
                  MyNormals[ 3*IndexNr[k] +1] = smoothNormal[1];
                a = MyNormals[ 3*IndexNr[k]    ]; // X-Value
                b = MyNormals[ 3*IndexNr[k] + 2]; // Z-Value
                if(b)
                  winkel = atan(a/b);      // Winkel zur Z-Achse
                else
		              winkel = sign(a)*M_PI_2;
                if( fabs(winkel - NVwinkel[2]) < creaseAngle )
		              MyNormals[ 3*IndexNr[k] + 2] = smoothNormal[2];
              }
            }   // mittlerer NV war vorhanden
          }   // es waren mehrere NV
        }   // Punkt war dabei
      }  // alle Inidizes
    }  // alle Punkte
  }  // creaseAngle
}

//---------------------------------------------------------------------
// Zeichnung vorbereiten
void FaceSetNode::PrepareForDraw()
{
  int      i;

  if(haveCoords && haveCoordIndex) {

    // die Normalenvektoren vorbereiten
    if(NormalPerVertex)
      myDrawing = drawNormalPerVertex;
    else
      myDrawing = drawNormalPerFace;
    whichDrawing = (int)myDrawing;// spezielle Zeichenmethode ermoeglichen

    if(!haveNormals) 		 // Normalenvektoren muessen berechnet werden
      CalcNormals();

    // die Farben vorbereiten
    if(haveColors) {		// Farben sind vorgegeben
      if(!ColorPerVertex) {	// Farben pro Flaeche
      	myDrawing = drawColorPerFace;
	      if(!haveColorIndex) {	// Indizes sind nicht vorgegeben
          #ifndef  USE_STATIC_FIELDS
	        MyColorIndex = NULL;
          MyColorIndex   = new int[MyCoordIndexSize]; // Groesse sollte reichen
          #endif
          if(MyColorIndex != NULL) {
	          MyColorIndexSize = MyCoordIndexSize;
	          for(i = 0; i < MyColorIndexSize; i++)
	            MyColorIndex[i] = i;// Farbwert nach Reihenfolge waehlen
            whichDrawing += (int)myDrawing; // fuer optimale Zeichenmethode
          }
        }
	      else
	        whichDrawing += (int)myDrawing; // fuer optimale Zeichenmethode
      }
      else {
      	myDrawing = drawColorPerVertex; // Farben pro Eckpunkt
	      if(!haveColorIndex) {	// Indizes sind nicht vorgegeben
          #ifndef  USE_STATIC_FIELDS
	        MyColorIndex = MyCoordIndex;
          #else
          memcpy( MyColorIndex, MyCoordIndex, MyCoordIndexSize*sizeof(int) );
          #endif
        }
	      whichDrawing += (int)myDrawing; // fuer optimale Zeichenmethode
      }
    }

    // die Texturen vorbereiten
    if(haveTextureCoords) {
      myDrawing = drawTexture;
      whichDrawing += (int)myDrawing;
      if(!haveTextureIndex) {
        #ifndef  USE_STATIC_FIELDS
	      MyTextureIndex = MyCoordIndex;
        #else
        memcpy( MyTextureIndex, MyCoordIndex, MyCoordIndexSize*sizeof(int) );
        #endif
      }
    }
  }
}

//---------------------------------------------------------------------
// konkave Polygone aufsplitten und zeichnen
void FaceSetNode::DrawTesselated()
{
  int        i, j;
  GLdouble   myValues[11]; // entaelt in dieser Reihenfolge:
  			   // Vertex, Normals, Color, Textur
  GLUtriangulatorObj   *TessObj = NULL;

  if(!CCW)
    glFrontFace(GL_CW);
  if(haveAppearance)
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
  else
    glColor3f(0.8, 0.8, 0.8);

  TessObj = gluNewTess();
  if(TessObj) {

    gluTessCallback( TessObj, GLU_BEGIN,
                     (GLvoid (CALLBACK*) ()) &BeginNewPolygon );
    gluTessCallback( TessObj, GLU_END,
                     (GLvoid (CALLBACK*) ()) &DrawNewPolygon );

    i = j = 0;
    switch(whichDrawing) {
    // drawNormalPerVertex
    case 1:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexCall );
      while( i < MyCoordIndexSize ) {
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[3] = (GLdouble)(MyNormals[3*i   ]);
          myValues[4] = (GLdouble)(MyNormals[3*i +1]);
          myValues[5] = (GLdouble)(MyNormals[3*i +2]);
          #ifdef GLU_VERSION_1_2
          gluTessNormal( TessObj, MyNormals[3*i],
          		           MyNormals[3*i +1], MyNormals[3*i +2] );
          #endif // GLU_VERSION_1_2
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;
    // drawNormalPerFace
    case 2:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexCall );
      while( i < MyCoordIndexSize ) {
        myValues[3] = (GLdouble)(MyNormals[3*i   ]);
        myValues[4] = (GLdouble)(MyNormals[3*i +1]);
        myValues[5] = (GLdouble)(MyNormals[3*i +2]);
        #ifdef GLU_VERSION_1_2
        gluTessNormal( TessObj, MyNormals[3*i],
              	       MyNormals[3*i +1], MyNormals[3*i +2] );
        #endif // GLU_VERSION_1_2
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerVertex mit Texturen
    case 6:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) () ) &VertexAndTextureCall );
      while( i < MyCoordIndexSize ) {
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[3]  = (GLdouble)(MyNormals[3*i   ]);
          myValues[4]  = (GLdouble)(MyNormals[3*i +1]);
          myValues[5]  = (GLdouble)(MyNormals[3*i +2]);
          myValues[9]  = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i]]);
          myValues[10] = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i] +1]);
          #ifdef GLU_VERSION_1_2
          gluTessNormal( TessObj, MyNormals[3*i],
                    		 MyNormals[3*i +1], MyNormals[3*i +2] );
          #endif // GLU_VERSION_1_2
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;
      // drawNormalPerFace mit Texturen
    case 7:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexAndTextureCall );
      while( i < MyCoordIndexSize ) {
        myValues[3] = (GLdouble)(MyNormals[3*i   ]);
        myValues[4] = (GLdouble)(MyNormals[3*i +1]);
        myValues[5] = (GLdouble)(MyNormals[3*i +2]);
        #ifdef GLU_VERSION_1_2
        gluTessNormal( TessObj, MyNormals[3*i],
              	       MyNormals[3*i +1], MyNormals[3*i +2] );
        #endif // GLU_VERSION_1_2
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[9]  = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i]]);
          myValues[10] = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i] +1]);
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerVertex + drawColorPerVertex
    case 11:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexAndColorCall );
      while( i < MyCoordIndexSize ) {
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[3] = (GLdouble)(MyNormals[3*i   ]);
          myValues[4] = (GLdouble)(MyNormals[3*i +1]);
          myValues[5] = (GLdouble)(MyNormals[3*i +2]);
          myValues[6] = (GLdouble)(MyColors[3*MyColorIndex[i]   ]);
          myValues[7] = (GLdouble)(MyColors[3*MyColorIndex[i] +1]);
          myValues[8] = (GLdouble)(MyColors[3*MyColorIndex[i] +2]);
          #ifdef GLU_VERSION_1_2
          gluTessNormal( TessObj, MyNormals[3*i],
                    		 MyNormals[3*i +1], MyNormals[3*i +2] );
          #endif
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;
    // drawNormalPerFace + drawColorPerVertex
    case 12:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexAndColorCall );
      while( i < MyCoordIndexSize ) {
        myValues[3] = (GLdouble)(MyNormals[3*i   ]);
        myValues[4] = (GLdouble)(MyNormals[3*i +1]);
        myValues[5] = (GLdouble)(MyNormals[3*i +2]);
        #ifdef GLU_VERSION_1_2
        gluTessNormal( TessObj, MyNormals[3*i],
              	       MyNormals[3*i +1], MyNormals[3*i +2] );
        #endif // GLU_VERSION_1_2
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[6] = (GLdouble)(MyColors[3*MyColorIndex[i]   ]);
          myValues[7] = (GLdouble)(MyColors[3*MyColorIndex[i] +1]);
          myValues[8] = (GLdouble)(MyColors[3*MyColorIndex[i] +2]);
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerVertex + drawColorPerVertex + Texturen
    case 16:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &CompleteValueCall );
      while( i < MyCoordIndexSize ) {
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[3]  = (GLdouble)(MyNormals[3*i   ]);
          myValues[4]  = (GLdouble)(MyNormals[3*i +1]);
          myValues[5]  = (GLdouble)(MyNormals[3*i +2]);
          myValues[6]  = (GLdouble)(MyColors[3*MyColorIndex[i]   ]);
          myValues[7]  = (GLdouble)(MyColors[3*MyColorIndex[i] +1]);
          myValues[8]  = (GLdouble)(MyColors[3*MyColorIndex[i] +2]);
          myValues[9]  = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i]]);
          myValues[10] = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i] +1]);
          #ifdef GLU_VERSION_1_2
          gluTessNormal( TessObj, MyNormals[3*i],
                    		 MyNormals[3*i +1], MyNormals[3*i +2] );
          #endif // GLU_VERSION_1_2
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerFace + drawColorPerVertex + Texturen
    case 17:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &CompleteValueCall );
      while( i < MyCoordIndexSize ) {
        myValues[3] = (GLdouble)(MyNormals[3*i   ]);
        myValues[4] = (GLdouble)(MyNormals[3*i +1]);
        myValues[5] = (GLdouble)(MyNormals[3*i +2]);
        #ifdef GLU_VERSION_1_2
        gluTessNormal( TessObj, MyNormals[3*i],
              	       MyNormals[3*i +1], MyNormals[3*i +2] );
        #endif
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {
          myValues[0]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[6]  = (GLdouble)(MyColors[3*MyColorIndex[i]   ]);
          myValues[7]  = (GLdouble)(MyColors[3*MyColorIndex[i] +1]);
          myValues[8]  = (GLdouble)(MyColors[3*MyColorIndex[i] +2]);
          myValues[9]  = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i]]);
          myValues[10] = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i] +1]);
          gluTessVertex( TessObj, myValues, (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerVertex + drawColorPerFace
    case 21:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexCall );
      while( i < MyCoordIndexSize ) {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
        j++;
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[3] = (GLdouble)(MyNormals[3*i   ]);
          myValues[4] = (GLdouble)(MyNormals[3*i +1]);
          myValues[5] = (GLdouble)(MyNormals[3*i +2]);
          #ifdef GLU_VERSION_1_2
          gluTessNormal( TessObj, MyNormals[3*i],
                    		 MyNormals[3*i +1], MyNormals[3*i +2] );
          #endif // GLU_VERSION_1_2
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerFace + drawColorPerFace
    case 22:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexCall );
      while( i < MyCoordIndexSize ) {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
        myValues[3] = (GLdouble)(MyNormals[3*i   ]);
        myValues[4] = (GLdouble)(MyNormals[3*i +1]);
        myValues[5] = (GLdouble)(MyNormals[3*i +2]);
        #ifdef GLU_VERSION_1_2
        gluTessNormal( TessObj, MyNormals[3*i],
              	       MyNormals[3*i +1], MyNormals[3*i +2] );
        #endif 
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
	        i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerVertex + drawColorPerFace + Texturen
    case 26:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexCall );
      while( i < MyCoordIndexSize ) {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
        j++;
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2]  = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[3]  = (GLdouble)(MyNormals[3*i   ]);
          myValues[4]  = (GLdouble)(MyNormals[3*i +1]);
          myValues[5]  = (GLdouble)(MyNormals[3*i +2]);
          myValues[9]  = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i]]);
          myValues[10] = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i] +1]);
          #ifdef GLU_VERSION_1_2
          gluTessNormal( TessObj, MyNormals[3*i],
                    		 MyNormals[3*i +1], MyNormals[3*i +2] );
          #endif
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;

    // drawNormalPerFace + drawColorPerFace + Texturen
    case 27:
      gluTessCallback( TessObj, GLU_VERTEX,
                       (GLvoid (CALLBACK*) ()) &VertexCall );
      while( i < MyCoordIndexSize ) {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
        myValues[3] = (GLdouble)(MyNormals[3*i   ]);
        myValues[4] = (GLdouble)(MyNormals[3*i +1]);
        myValues[5] = (GLdouble)(MyNormals[3*i +2]);
        #ifdef GLU_VERSION_1_2
        gluTessNormal( TessObj, MyNormals[3*i],
              	       MyNormals[3*i +1], MyNormals[3*i +2] );
        #endif
        gluBeginPolygon(TessObj);
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          myValues[0] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i]   ]);
          myValues[1] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +1]);
          myValues[2] = (GLdouble)(MyCoords[ 3*MyCoordIndex[i] +2]);
          myValues[9]  = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i]]);
          myValues[10] = (GLdouble)(MyTextureCoords[2*MyTextureIndex[i] +1]);
          gluTessVertex( TessObj,
                         myValues,
                         (MyCoords + 3*MyCoordIndex[i]) );
          i++;				     // Polygon aufbauen
        }
        gluEndPolygon(TessObj);
        i++;      			     // das ist die -1
      }
      break;


    }
    glFrontFace(GL_CCW);
    gluDeleteTess(TessObj);
  }
}

//---------------------------------------------------------------------
// konvexe Polygone zeichnen
void FaceSetNode::DrawConvexOnly()
{
  int        i, j;

  if(!CCW)
    glFrontFace(GL_CW);

  if(haveAppearance)
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
  else
    glColor3f(0.8, 0.8, 0.8);

  i = j = 0;
  switch(whichDrawing) {
  // drawNormalPerVertex
  case 1:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          glNormal3fv( MyNormals + 3*i );
	        glVertex3fv( MyCoords  + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
        }
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;
    // drawNormalPerFace
  case 2:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glNormal3fv( MyNormals + 3*i );
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
          glVertex3fv( MyCoords  + 3*MyCoordIndex[i] );
          i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;

    // drawNormalPerVertex mit Texturen
  case 6:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glTexCoord2fv( MyTextureCoords + 2*MyTextureIndex[i] );
	        glNormal3fv( MyNormals + 3*i );
	        glVertex3fv( MyCoords  + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    glDisable(GL_TEXTURE_2D);
    break;
    // drawNormalPerFace mit Texturen
  case 7:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glNormal3fv( MyNormals + 3*i );
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glTexCoord2fv( MyTextureCoords + 2*MyTextureIndex[i] );
	        glVertex3fv( MyCoords  + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    glDisable(GL_TEXTURE_2D);
    break;

    // drawNormalPerVertex + drawColorPerVertex
  case 11:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glColor3fv( MyColors + 3*MyColorIndex[i] );
	        glNormal3fv( MyNormals + 3*i );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;
    // drawNormalPerFace + drawColorPerVertex
  case 12:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glNormal3fv( MyNormals + 3*i );
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glColor3fv( MyColors + 3*MyColorIndex[i] );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;

    // drawNormalPerVertex + drawColorPerVertex + Texturen
  case 16:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glColor3fv( MyColors + 3*MyColorIndex[i] );
	        glTexCoord2fv( MyTextureCoords + 2*MyTextureIndex[i] );
	        glNormal3fv( MyNormals + 3*i );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    glDisable(GL_TEXTURE_2D);
    break;
    // drawNormalPerFace + drawColorPerVertex + Texturen
  case 17:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glNormal3fv( MyNormals + 3*i );
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glColor3fv( MyColors + 3*MyColorIndex[i] );
	        glTexCoord2fv( MyTextureCoords + 2*MyTextureIndex[i] );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    glDisable(GL_TEXTURE_2D);
    break;

    // drawNormalPerVertex + drawColorPerFace
  case 21:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
	      j++;
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glNormal3fv( MyNormals + 3*i );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;
    // drawNormalPerFace + drawColorPerFace
  case 22:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
        glNormal3fv( MyNormals + 3*i );
	      j++;
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;

    // drawNormalPerVertex + drawColorPerFace + Texturen
  case 26:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
	      j++;
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glTexCoord2fv( MyTextureCoords + 2*MyTextureIndex[i] );
	        glNormal3fv( MyNormals + 3*i );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    glDisable(GL_TEXTURE_2D);
    break;
    // drawNormalPerFace + drawColorPerFace + Texturen
  case 27:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_POLYGON);
      {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
        glNormal3fv( MyNormals + 3*i );
	      j++;
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glTexCoord2fv( MyTextureCoords + 2*MyTextureIndex[i] );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    glDisable(GL_TEXTURE_2D);
    break;
  }

  glFrontFace(GL_CCW);
}


//=====================================================================
// Von einem VRML-Object abgeleitete LineSet-Node.
// Diese Klasse beschreibt einen Linienzug.
//=====================================================================
//
// Konstruktor
LineSetNode::LineSetNode()
{
  myDrawing	       = nothingToDraw;
  whichDrawing     = (int)myDrawing;
#ifdef USE_STATIC_FIELDS
   MyCoordIndexSize  = MyColorIndexSize   = INDEX_MULT*MAXVERTICES;
#else
  MyCoordIndex = MyColorIndex = NULL;
  MyCoordIndexSize = MyColorIndexSize   = 0;
#endif
  haveCoords       = haveColors = false;
  haveCoordIndex   = haveColorIndex    = false;
  ColorPerVertex   = true;

  myNodeType       = IndexedLineSet;
}

//---------------------------------------------------------------------
// Destruktor
LineSetNode::~LineSetNode()
{
#ifndef  USE_STATIC_FIELDS
  if( MyCoordIndex != NULL )
    delete MyCoordIndex;
  if( MyColorIndex != NULL )
    delete MyColorIndex;
#endif
}

//---------------------------------------------------------------------
// Zeiger auf Koordinatenfeld
void LineSetNode::AddCoordField(GLfloat *CoordField)
{
  MyCoords   = CoordField;
  haveCoords = true;
}

//---------------------------------------------------------------------
// Indizes auf Koordinatenfeld
void LineSetNode::AddCoordIndex(int *CoordIndexPtr, int CoordIndexSize)
{
#ifdef  USE_STATIC_FIELDS
  memcpy( MyCoordIndex, CoordIndexPtr, CoordIndexSize*sizeof(int) );
  MyCoordIndexSize = CoordIndexSize;
  haveCoordIndex   = true;
#else
  MyCoordIndex = NULL;
  MyCoordIndex = new int[CoordIndexSize];
  if(MyCoordIndex != NULL) {
    memcpy( MyCoordIndex, CoordIndexPtr, CoordIndexSize*sizeof(int) );
    MyCoordIndexSize = CoordIndexSize;
    haveCoordIndex   = true;
  }
#endif
}

//---------------------------------------------------------------------
// Zeiger auf ColorFeld
void LineSetNode::AddColorField(GLfloat *ColorField)
{
  MyColors   = ColorField;
  haveColors = true;
}

//---------------------------------------------------------------------
// Indizes auf ColorFeld
void LineSetNode::AddColorIndex(int *ColorIndex, int ColorIndexSize)
{
#ifdef  USE_STATIC_FIELDS
  memcpy( MyColorIndex, ColorIndex, ColorIndexSize*sizeof(int) );
  MyColorIndexSize = ColorIndexSize;
  haveColorIndex   = true;
#else
  MyColorIndex = NULL;
  MyColorIndex = new int[ColorIndexSize];
  if(MyColorIndex != NULL) {
    memcpy( MyColorIndex, ColorIndex, ColorIndexSize*sizeof(int) );
    MyColorIndexSize = ColorIndexSize;
    haveColorIndex   = true;
  }
#endif
}

//---------------------------------------------------------------------
// Zeichnung vorbereiten
void LineSetNode::PrepareForDraw()
{
  int      i;

  if(haveCoords && haveCoordIndex) {
    whichDrawing = (int)myDrawing;// spezielle Zeichenmethode ermoeglichen
    // die Farben vorbereiten
    if(haveColors) {		// Farben sind vorgegeben
      if(!ColorPerVertex) {	// Farben pro Flaeche
      	myDrawing = drawColorPerFace;
	      if(!haveColorIndex) {	// Indizes sind nicht vorgegeben
          #ifndef  USE_STATIC_FIELDS
	        MyColorIndex = NULL;
          MyColorIndex   = new int[MyCoordIndexSize]; // Groesse sollte reichen
          #endif
          if(MyColorIndex != NULL) {
	          MyColorIndexSize = MyCoordIndexSize;
	          for(i = 0; i < MyColorIndexSize; i++)
	            MyColorIndex[i] = i;// Farbwert nach Reihenfolge waehlen
            whichDrawing += (int)myDrawing; // fuer optimale Zeichenmethode
          }
        }
	      else
	        whichDrawing += (int)myDrawing; // fuer optimale Zeichenmethode
      }
      else {
      	myDrawing = drawColorPerVertex; // Farben pro Eckpunkt
	      if(!haveColorIndex) {	// Indizes sind nicht vorgegeben
          #ifndef  USE_STATIC_FIELDS
	        MyColorIndex = MyCoordIndex;
          #else
          memcpy( MyColorIndex, MyCoordIndex, MyCoordIndexSize*sizeof(int) );
          #endif
        }
	      whichDrawing += (int)myDrawing; // fuer optimale Zeichenmethode
      }
    }
  }
}

//---------------------------------------------------------------------
// konvexe Polygone zeichnen
void LineSetNode::DrawConvexOnly()
{
  int        i, j;
  
  if(haveAppearance)
    MyAppearance->DrawNode();            // vorgegebene Materialeigenschaften
  else
    glColor3f(0.8, 0.8, 0.8);

  i = j = 0;
  switch(whichDrawing) {
  // drawNormalPerFace + drawColorPerFace
  case  0:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_LINE_STRIP); // GL_LINE_STRIP GL_POLYGON
      {
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;
  // drawColorPerVertex
  case 10:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_LINE_STRIP); /// GL_LINE_STRIP GL_POLYGON
      {
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glColor3fv( MyColors + 3*MyColorIndex[i] );
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;
  case 20:
    while( i < MyCoordIndexSize ) {
      glBegin(GL_LINE_STRIP);
      {
        glColor3fv( MyColors + 3*MyColorIndex[j] );
	      j++;
        while( (MyCoordIndex[i] >= 0) && (i < MyCoordIndexSize) ) {// -1 ist Abschluss
	        glVertex3fv( MyCoords + 3*MyCoordIndex[i] );
	        i++;				     // Polygon aufbauen
      	}
      }
      glEnd();
      i++;      			     // das ist die -1
    }
    break;
  }
  glFrontFace(GL_CCW);
}


//=====================================================================
// nicht durch den Standard definierte Ergaenzungen

//=====================================================================
//  Klasse zeichnet ein Koordinatensystem und dient als
//  Ersatz fuer nicht vorhandene Beschreibungsdateien (deshalb eigene Klasse)
//=====================================================================
//
// Konstruktor
GLCoordSystem::GLCoordSystem()
{
  for(int i = 0; i < 3; i++)
    CoordV0[i] = CoordV1[i] = 0.0;
  CoordV1[0] =  5.0; // Laenge einer Achse
  CoordA1[0] = -0.3; CoordA1[1] = 0.05; CoordA1[1] = -0.05; // Pfeilspitzen
  CoordA2[0] = -0.3; CoordA2[1] = 0.05; CoordA2[1] =  0.05;
  CoordList  = false;
}

//---------------------------------------------------------------------
// Coordsystem ist auf Fahrzeug ausgerichtet
void GLCoordSystem::DrawGLObjectAsList()
{
  if( !CoordList ) {
    CoordListNr = glGenLists(1); // neue Listenummer erzeugen
    glNewList(CoordListNr, GL_COMPILE_AND_EXECUTE); // enthaelt alle Teilmodelle

    int i;

    for(i = 0; i < 3; i++) {
      glPushMatrix();
      switch(i) {
      case 0: glColor3f(1.0, 0.0, 0.0);
      	      break;
      case 1: glColor3f(0.0, 1.0, 0.0);
      	      glRotatef(90.0, 0.0, 0.0, 1.0);
      	      break; // Z-Achse
      case 2: glColor3f(0.0, 0.0, 1.0);
      	      glRotatef(-90.0, 0.0, 1.0, 0.0);
              break;  // Y-Achse
      }
      glBegin(GL_LINES);
      {
        glVertex3fv(CoordV0);
        glVertex3fv(CoordV1);
      }
      glEnd();

      glBegin(GL_TRIANGLES);
      {
        glVertex3f(0.9, -0.1, -0.1);
        glVertex3f(1.1, -0.1,  0.1);
	      glVertex3f(1.0,  0.1,  0.0);
        glVertex3f(1.9, -0.1, -0.1);
        glVertex3f(2.1, -0.1,  0.1);
	      glVertex3f(2.0,  0.1,  0.0);
        glVertex3f(2.9, -0.1, -0.1);
        glVertex3f(3.1, -0.1,  0.1);
	      glVertex3f(3.0,  0.1,  0.0);
        glVertex3f(3.9, -0.1, -0.1);
        glVertex3f(4.1, -0.1,  0.1);
	      glVertex3f(4.0,  0.1,  0.0);
      }
      glEnd();
      /*
      glTranslatef(4.0, 0.0, 0.0);
      for(j = 0; j < 8; j++) {
	      glBegin(GL_POLYGON);
        {
          glVertex3fv(CoordA1);
          glVertex3fv(CoordA2);
          glVertex3fv(CoordV0); // 0,0,0
          glVertex3fv(CoordA1);
        }
	      glEnd();
	      glRotatef(45.0, 1.0, 0.0, 0.0);
      }
      */
      glPopMatrix();
    }

    glEndList();
    CoordList = true;
  }
  else
    glCallList(CoordListNr);
}

